@@ -1,19 +1,20 @@
 import argparse
+import bz2
+import hashlib
+import json
 import logging
+import multiprocessing
 import os
 import pdb
 import shutil
 import sys
-import json
 import tarfile
 import tempfile
-import traceback
-from glob import fnmatch
+import fnmatch
 from pprint import pformat
-import bz2
+
 import requests
 import yaml
-import hashlib
 
 logger = None
 
@@ -27,8 +28,9 @@ DEFAULT_PLATFORMS = ['linux-64',
 
 
 def _maybe_split_channel(channel):
-    """Split channel if it is fully qualified. Otherwise default to
-    conda.anaconda.org
+    """Split channel if it is fully qualified.
+
+    Otherwise default to conda.anaconda.org
 
     Parameters
     ----------
@@ -39,13 +41,15 @@ def _maybe_split_channel(channel):
     Returns
     -------
     download_template : str
-        defaults to "https://conda.anaconda.org/{channel}/{platform}/{file_name}"
+        defaults to
+        "https://conda.anaconda.org/{channel}/{platform}/{file_name}"
         The base url will be modified if the `channel` input parameter is
         fully qualified
     channel : str
         The name-only channel. If the channel input param is something like
         "conda-forge", then "conda-forge" will be returned. If the channel
         input param is something like "https://repo.continuum.io/pkgs/free/"
+
     """
     # strip trailing slashes
     channel = channel.strip('/')
@@ -60,7 +64,8 @@ def _maybe_split_channel(channel):
     # looks like we are being given a fully qualified channel
     download_base, channel = channel.rsplit('/', 1)
     download_template = download_base + url_suffix
-    logger.debug('download_template=%s. channel=%s', download_template, channel)
+    logger.debug('download_template=%s. channel=%s',
+                 download_template, channel)
     return download_template, channel
 
 
@@ -79,6 +84,7 @@ def _match(all_packages, key_glob_dict):
     matched : dict
         Iterable of package metadata dicts which match the `target_packages`
         (key, glob_value) tuples
+
     """
     matched = dict()
     key_glob_dict = {key.lower(): glob.lower()
@@ -108,7 +114,8 @@ def _make_arg_parser():
     argument_parser : argparse.ArgumentParser
         The instantiated argument parser for this CLI
     """
-    ap = argparse.ArgumentParser(description="CLI interface for conda-mirror.py")
+    ap = argparse.ArgumentParser(
+        description="CLI interface for conda-mirror.py")
 
     ap.add_argument(
         '--upstream-channel',
@@ -122,10 +129,11 @@ def _make_arg_parser():
     )
     ap.add_argument(
         '--temp-directory',
-        help=('Temporary download location for the packages. Defaults to a '
-              'randomly selected temporary directory. Note that you might need '
-              'to specify a different location if your default temp directory '
-              'has less available space than your mirroring target'),
+        help=(
+            'Temporary download location for the packages. Defaults to a '
+            'randomly selected temporary directory. Note that you might need '
+            'to specify a different location if your default temp directory '
+            'has less available space than your mirroring target'),
         default=tempfile.gettempdir()
     )
     ap.add_argument(
@@ -151,6 +159,13 @@ def _make_arg_parser():
         help="Enable PDB debugging on exception",
         default=False,
     )
+    ap.add_argument(
+        '--num-threads',
+        action="store",
+        default=1,
+        type=int,
+        help="Num of threads for validation. 1: Serial mode. 0: All available."
+        )
     ap.add_argument(
         '--version',
         action="store_true",
@@ -186,7 +201,7 @@ def _init_logger(verbosity):
           file=sys.stdout)
 
 
-def cli():
+def _parse_and_format_args():
     """
     Collect arguments from sys.argv and invoke the main() function.
     """
@@ -199,12 +214,12 @@ def cli():
     if args.version:
         from . import __version__
         print(__version__)
-        return
+        sys.exit(1)
 
     for required in ('target_directory', 'platform', 'upstream_channel'):
         if not getattr(args, required):
-            logger.error("Missing required argument: %s", required)
-            return
+            raise ValueError("Missing command line argument: %s", required)
+
     if args.pdb:
         # set the pdb_hook as the except hook for all exceptions
         def pdb_hook(exctype, value, traceback):
@@ -220,8 +235,21 @@ def cli():
     blacklist = config_dict.get('blacklist')
     whitelist = config_dict.get('whitelist')
 
-    main(args.upstream_channel, args.target_directory, args.temp_directory,
-         args.platform, blacklist, whitelist)
+    return {
+        'upstream_channel': args.upstream_channel,
+        'target_directory': args.target_directory,
+        'temp_directory': args.temp_directory,
+        'platform': args.platform,
+        'num_threads': args.num_threads,
+        'blacklist': blacklist,
+        'whitelist': whitelist
+    }
+
+
+def cli():
+    """Thin wrapper around parsing the cli args and calling main with them
+    """
+    main(**_parse_and_format_args())
 
 
 def _remove_package(pkg_path, reason):
@@ -232,11 +260,18 @@ def _remove_package(pkg_path, reason):
     ----------
     pkg_path : str
         Path to a conda package that should be removed
-    """
-    msg = "Removing: %s. Reason: %s"
-    logger.warning(msg, pkg_path, reason)
-    os.remove(pkg_path)
 
+    Returns
+    -------
+    pkg_path : str
+        The full path to the package that is being removed
+    reason : str
+        The reason why the package is being removed
+    """
+    msg = "Removing: %s. Reason: %s" % (pkg_path, reason)
+    logger.warning(msg)
+    os.remove(pkg_path)
+    return pkg_path, msg
 
 
 def _validate(filename, md5=None, size=None):
@@ -255,6 +290,13 @@ def _validate(filename, md5=None, size=None):
     size : int, optional
         if provided, stat the file at `filename` and make sure its size
         matches `size`
+
+    Returns
+    -------
+    pkg_path : str
+        The full path to the package that is being removed
+    reason : str
+        The reason why the package is being removed
     """
     try:
         t = tarfile.open(filename)
@@ -262,19 +304,19 @@ def _validate(filename, md5=None, size=None):
     except tarfile.TarError:
         logger.info("Validation failed because conda package is corrupted.",
                     exc_info=True)
-        _remove_package(filename, reason="Tarfile read failure")
-        return
+        return _remove_package(filename, reason="Tarfile read failure")
     if size:
         if os.stat(filename).st_size != size:
-            _remove_package(filename, reason="Failed size test")
-            return
+            return _remove_package(filename, reason="Failed size test")
     if md5:
         calc = hashlib.md5(open(filename, 'rb').read()).hexdigest()
         if calc != md5:
-            _remove_package(
+            return _remove_package(
                 filename,
-                reason="Failed md5 validation. Expected: %s. Computed: %s" % (calc, md5))
-            return
+                reason="Failed md5 validation. Expected: %s. Computed: %s"
+                % (calc, md5))
+
+    return filename, None
 
 
 def get_repodata(channel, platform):
@@ -339,13 +381,14 @@ def _list_conda_packages(local_dir):
     return fnmatch.filter(contents, "*.tar.bz2")
 
 
-def _validate_packages(package_repodata, package_directory):
+def _validate_packages(package_repodata, package_directory, num_threads=1):
     """Validate local conda packages.
 
-    NOTE: This is slow.
-    NOTE2: This will remove any packages that are in `package_directory` that
+    NOTE1: This will remove any packages that are in `package_directory` that
            are not in `repodata` and also any packages that fail the package
            validation
+    NOTE2: In concurrent mode (num_threads is not 1) this might be hard to kill
+           using CTRL-C.
 
     Parameters
     ----------
@@ -353,31 +396,98 @@ def _validate_packages(package_repodata, package_directory):
         The contents of repodata.json
     package_directory : str
         Path to the local repo that contains conda packages
+    num_threads : int
+        Number of concurrent processes to use. Set to `0` to use a number of
+        processes equal to the number of cores in the system. Defaults to `1`
+        (i.e. serial package validation).
+
+    Returns
+    -------
+    list
+        Iterable of twoples of (pkg_path, reason) where
+        pkg_path : str
+            The full path to the package that is being removed
+        reason : str
+            The reason why the package is being removed
     """
     # validate local conda packages
     local_packages = _list_conda_packages(package_directory)
-    for idx, package in enumerate(sorted(local_packages)):
-        # ensure the packages in this directory are in the upstream
-        # repodata.json
-        try:
-            package_metadata = package_repodata[package]
-        except KeyError:
-            logger.warning("%s is not in the upstream index. Removing...",
-                           package)
-            _remove_package(os.path.join(package_directory, package),
-                            reason="Package is not in the repodata index")
-        else:
-            # validate the integrity of the package, the size of the package and
-            # its hashes
-            logger.info('Validating %s. %s of %s', package, idx,
-                        len(local_packages))
-            _validate(os.path.join(package_directory, package),
-                      md5=package_metadata.get('md5'),
-                      size=package_metadata.get('size'))
+
+    # create argument list (necessary because multiprocessing.Pool.map does not
+    # accept additional args to be passed to the mapped function)
+    num_packages = len(local_packages)
+    val_func_arg_list = [(package, num, num_packages, package_repodata,
+                          package_directory)
+                         for num, package in enumerate(sorted(local_packages))]
+
+    if num_threads is 1 or num_threads is None:
+        # Do serial package validation (Takes a long time for large repos)
+        validation_results = map(_validate_or_remove_package,
+                                 val_func_arg_list)
+    else:
+        if num_threads is 0:
+            num_threads = os.cpu_count()
+            logger.debug('num_threads=0 so it will be replaced by all available '
+                         'cores: %s' % num_threads)
+        logger.info('Will use {} threads for package validation.'
+                    ''.format(num_threads))
+        p = multiprocessing.Pool(num_threads)
+        validation_results = p.map(_validate_or_remove_package,
+                                   val_func_arg_list)
+        p.close()
+        p.join()
+
+    return validation_results
+
+
+def _validate_or_remove_package(args):
+    """Validata or remove package.
+
+    Parameters
+    ----------
+    args : tuple
+        - `args[0]` is `package`.
+        - `args[1]` is the number of the package in the list of all packages.
+        - `args[2]` is the number of all packages.
+        - `args[3]` is `package_repodata`.
+        - `args[4]` is `package_directory`.
+
+    Returns
+    -------
+    pkg_path : str
+        The full path to the package that is being removed
+    reason : str
+        The reason why the package is being removed
+    """
+    # unpack arg tuple tuple
+    package = args[0]
+    num = args[1]
+    num_packages = args[2]
+    package_repodata = args[3]
+    package_directory = args[4]
+
+    # ensure the packages in this directory are in the upstream
+    # repodata.json
+    try:
+        package_metadata = package_repodata[package]
+    except KeyError:
+        logger.warning("%s is not in the upstream index. Removing...",
+                       package)
+        reason = "Package is not in the repodata index"
+        package_path = os.path.join(package_directory, package)
+        return _remove_package(package_path, reason=reason)
+    # validate the integrity of the package, the size of the package and
+    # its hashes
+    logger.info('Validating {:4d} of {:4d}: {}.'.format(num, num_packages,
+                                                        package))
+    package_path = os.path.join(package_directory, package)
+    return _validate(package_path,
+                     md5=package_metadata.get('md5'),
+                     size=package_metadata.get('size'))
 
 
 def main(upstream_channel, target_directory, temp_directory, platform,
-         blacklist=None, whitelist=None):
+         blacklist=None, whitelist=None, num_threads=1):
     """
 
     Parameters
@@ -406,6 +516,20 @@ def main(upstream_channel, target_directory, temp_directory, platform,
         The values of blacklist should be (key, glob) where key is one of the
         keys in the repodata['packages'] dicts and glob is a thing to match
         on.  Note that all comparisons will be laundered through lowercasing.
+    num_threads : int
+        Number of threads to be used for concurrent validation.  Defaults to
+        `num_threads=1` for non-concurrent mode.  To use all available cores,
+        set `num_threads=0`.
+
+    Returns
+    -------
+    dict
+        Summary of what was removed and what was downloaded.
+        keys are:
+        - validation : set of (path, reason) for each package that was validated.
+                       packages where reason=None is a sentinel for a successful validation
+        - download : set of (url, download_path) for each package that
+                     was downloaded
 
     Notes
     -----
@@ -445,7 +569,9 @@ def main(upstream_channel, target_directory, temp_directory, platform,
     # 7. copy new packages to repo directory
     # 8. download repodata.json and repodata.json.bz2
     # 9. copy new repodata.json and repodata.json.bz2 into the repo
-
+    summary = {'validating-existing': set(),
+               'validating-new': set(),
+               'downloaded': set()}
     # Implementation:
     if not os.path.exists(os.path.join(target_directory, platform)):
         os.makedirs(os.path.join(target_directory, platform))
@@ -456,43 +582,38 @@ def main(upstream_channel, target_directory, temp_directory, platform,
     # 1. validate local repo
     # validating all packages is taking many hours.
     # _validate_packages(repodata=repodata,
-    #                    package_directory=local_directory)
+    #                    package_directory=local_directory,
+    #                    num_threads=num_threads)
 
     # 2. figure out blacklisted packages
     blacklist_packages = {}
     whitelist_packages = {}
     # match blacklist conditions
     if blacklist:
-        logger.debug("blacklist")
         blacklist_packages = {}
         for blist in blacklist:
             matched_packages = _match(packages, blist)
             blacklist_packages.update(matched_packages)
-        logger.debug(pformat(sorted(blacklist_packages)))
 
     # 3. un-blacklist packages that are actually whitelisted
     # match whitelist on blacklist
     if whitelist:
-        logger.debug("whitelist")
         whitelist_packages = {}
         for wlist in whitelist:
             matched_packages = _match(packages, wlist)
             whitelist_packages.update(matched_packages)
-        logger.debug(pformat(sorted(whitelist_packages)))
     # make final mirror list of not-blacklist + whitelist
     true_blacklist = set(blacklist_packages.keys()) - set(
         whitelist_packages.keys())
-    logger.debug('true blacklist')
-    logger.debug(pformat(sorted(whitelist_packages)))
     possible_packages_to_mirror = set(packages.keys()) - true_blacklist
-    logger.debug('possible_packages_to_mirror')
-    logger.debug(pformat(sorted(possible_packages_to_mirror)))
 
     # 4. Validate all local packages
     # construct the desired package repodata
-    desired_repodata = {pkgname: packages[pkgname] for pkgname in possible_packages_to_mirror}
-    _validate_packages(desired_repodata, local_directory)
+    desired_repodata = {pkgname: packages[pkgname]
+                        for pkgname in possible_packages_to_mirror}
 
+    validation_results = _validate_packages(desired_repodata, local_directory, num_threads)
+    summary['validating-existing'].update(validation_results)
     # 5. figure out final list of packages to mirror
     # do the set difference of what is local and what is in the final
     # mirror list
@@ -515,18 +636,20 @@ def main(upstream_channel, target_directory, temp_directory, platform,
                 platform=platform,
                 file_name=package_name)
             _download(url, download_dir)
+            summary['downloaded'].add((url, download_dir))
 
         # validate all packages in the download directory
-        _validate_packages(packages, download_dir)
-        logger.debug('contents of %s are %s',
+        validation_results = _validate_packages(packages, download_dir,
+                                                num_threads=num_threads)
+        summary['validating-new'].update(validation_results)
+        logger.debug('Newly downloaded files at %s are %s',
                      download_dir,
                      pformat(os.listdir(download_dir)))
 
         # 8. Use already downloaded repodata.json contents but prune it of
         # packages we don't want
-        repodata_path = os.path.join(download_dir, 'repodata.json')
-
         repodata = {'info': info, 'packages': packages}
+
         # compute the packages that we have locally
         packages_we_have = set(local_packages +
                                _list_conda_packages(download_dir))
@@ -556,6 +679,8 @@ def main(upstream_channel, target_directory, temp_directory, platform,
         noarch_repodata = {'info': {}, 'packages': {}}
         _write_repodata(noarch_path, noarch_repodata)
 
+    return summary
+
 
 def _write_repodata(package_dir, repodata_dict):
     data = json.dumps(repodata_dict, indent=2, sort_keys=True)
